/************************************************************************
     File:        ArcBallCam.cpp
        substantially updated October 2007, an upgrade from MyArcBall

     Author:     
                  Michael Gleicher, gleicher@cs.wisc.edu
     Modifier
                  Yu-Chi Lai, yu-chi@cs.wisc.edu
     
     Comment:    
				Written October, 2001, Michael L. Gleicher
				Updated October, 2005 for CS559 - Michael L. Gleicher

     See the header file for information.

     CVS Info: Date checked in: $Date: 2008/10/08 16:25:33 $
               Revision number: $Revision: 1.4 $
   
     Platform:    Visio Studio.Net 2005
*************************************************************************/

#include "ArcBallCam.H"
#include <math.h>

#include <windows.h>

// the FlTk headers have lots of warnings - these are bad, but there's not
// much we can do about them
#pragma warning(push)
#pragma warning(disable:4311)		// convert void* to long
#pragma warning(disable:4312)		// convert long to void*
#include <FL/Fl_Gl_Window.h>
#include <Fl/Fl.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <Fl/Fl_Double_Window.h>

#include "stdio.h"


//**************************************************************************
//
//==========================================================================
ArcBallCam::ArcBallCam() : wind(0),
						 fieldOfView(40),
						 eyeX(0), eyeY(0), eyeZ(20),
						 initEyeZ(20), mode(None), 
						 panX(0), panY(0)
{
}

// really do the setup stuff, once you have the window
void ArcBallCam::setup(Fl_Gl_Window* _wind, float _fieldOfView, float _eyeZ)
{
	wind = _wind;
	fieldOfView = _fieldOfView;
	eyeZ = _eyeZ;
	initEyeZ = _eyeZ;
	eyeX = 0;
	eyeY = 0;
}


void ArcBallCam::setProjection(bool doClear)
{
  glMatrixMode(GL_PROJECTION);
  glViewport(0,0,wind->w(),wind->h());
  if (doClear)
	  glLoadIdentity();

  // Compute the aspect ratio so we don't distort things
  double aspect = ((double) wind->w()) / ((double) wind->h());
  gluPerspective(fieldOfView, aspect, .1, 1000);

  // Put the camera where we want it to be
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Use the transformation in the ArcBall
  glTranslatef(-eyeX, -eyeY, -eyeZ);
  multMatrix();
}

int ArcBallCam::handle(int e)
{
	switch(e) {
		case FL_PUSH:
			mode = None;
			if (Fl::event_button() == 3) { // right mouse button down
					float x,y;
					getMouseNDC(x,y);
					down(x,y);
					wind->damage(1);
					mode = (Fl::event_state() & FL_ALT) ? Pan : Rotate;
					return 1;
			};
			break;
		case FL_RELEASE:
			if (mode != None) {
				wind->damage(1);
				mode = None;
				return 1;
			}
			break;
		case FL_DRAG: // if the user drags the mouse
			if(mode != None) { // we're taking the drags
				float x,y;
				getMouseNDC(x,y);
				computeNow(x,y);
				wind->damage(1);
				return 1;
			}
	}
	return 0;
}

//**************************************************************************
//
// Get the mouse in NDC
//==========================================================================
void ArcBallCam::getMouseNDC(float& x, float& y)
//==========================================================================
{
	// notice, we put everything into doubles so we can do the math
	float mx = (float) Fl::event_x();	// remeber where the mouse went down
	float my = (float) Fl::event_y();

	// we will assume that the viewport is the same as the window size
	float wd = (float) wind->w();
	float hd = (float) wind->h();

	// remember that FlTk has Y going the wrong way!
	my = hd-my;

	x = (mx / wd) * 2.0f - 1.f;
	y = (my / hd) * 2.0f - 1.f;
}

//**************************************************************************
//
// when the mouse goes down, remember where. also, clear out the now
// position by putting it into start
//==========================================================================
void ArcBallCam::down(const float x, const float y)
//==========================================================================
{
	start = now * start;
	now = Quat();		// identity

	downX = x;
	downY = y;

	panX = 0;
	panY = 0;
}

//**************************************************************************
//
// get the whole matrix!
//==========================================================================
void ArcBallCam::getMatrix(HMatrix m) const
//==========================================================================
{
	Quat qAll = now * start;
	qAll = qAll.conjugate();   // since Ken does everything transposed
	qAll.toMatrix(m);
}

//**************************************************************************
//
// a simplified interface - so you never see the insides of arcball
//==========================================================================
void ArcBallCam::multMatrix()
//==========================================================================
{
	HMatrix m;
	getMatrix(m);
    glMultMatrixf((float*) m);
}
//**************************************************************************
//
//==========================================================================
void ArcBallCam::reset()
//==========================================================================
{
	start.x = 0;
	start.y = 0;
	start.z = 0;
	start.w = 1;

	now.x = now.y = now.z = 0; now.w = 1;
}

//**************************************************************************
//
//==========================================================================
void ArcBallCam::spin(float x, float y, float z)
//==========================================================================
{
	// printf("\nstart was %g %g %g %g\n",start.x,start.y,start.z,start.w);
	// first, get rid of anything cached
	start = now * start;
	now = Quat();

	float iw = x*x + y*y + z*z;
	if (iw<1)
		iw = sqrt(1-iw);
	else
		iw = 0;
	Quat newq(x,y,z,iw);

	newq.renorm();
	start = newq * start;
	// printf(" (no norm) %g %g %g %g\n",start.x,start.y,start.z,start.w);
	start.renorm();
	// printf("start is %g %g %g %g\n",start.x,start.y,start.z,start.w);
}

//**************************************************************************
//
// the gory guts

//**************************************************************************
//
// a helper - figure out where in XYZ a mouse down goes
// assumes sphere of unit radius
//==========================================================================
static void onUnitSphere(const float mx, const float my,
													float& x, float& y, float& z)
//==========================================================================
{
	x = mx;		// should divide radius
	y = my;
	float mag = x*x + y*y;
	if (mag > 1.0f) {
		float scale = 1.0f / ((float) sqrt(mag));
		x *= scale;
		y *= scale;
		z = 0;
	} else {
		z = (float) sqrt(1 - mag);
	}
}

//**************************************************************************
//
//==========================================================================
void ArcBallCam::computeNow(const float nowX, const float nowY)
//==========================================================================
{
	if (mode==Rotate) {
		float dx,dy,dz;
		float mx,my,mz;
		onUnitSphere(downX, downY, dx, dy, dz);
		onUnitSphere(nowX, nowY, mx, my, mz);

		// here we compute the quaternion between these two points
		now.x = dy*mz - dz*my;
		now.y = dz*mx - dx*mz;
		now.z = dx*my - dy*mx;
		now.w = dx*mx + dy*my + dz*mz;

		now.renorm();		// just in case...
	} else if (mode==Pan) {
		float dx = (nowX-downX) * eyeZ;
		float dy = (nowY-downY) * eyeZ;

		eyeX += panX - dx;
		eyeY += panY - dy;
		panX = dx;
		panY = dy;
	}
}

// CVS Header - if you don't know what this is, don't worry about it
// $Header: /u/c/s/cs559-1/private/CVS/ConstructionSite/ConstructionSite/Utilities/ArcBallCam.cpp,v 1.4 2008/10/08 16:25:33 gleicher Exp $
